//===---------- CustomPatternComparator.h - Code pattern finder -----------===//
//
//       SimpLL - Program simplifier for analysis of semantic difference      //
//
// This file is published under Apache 2.0 license. See LICENSE for details.
// Author: Petr Silling, psilling@redhat.com
//===----------------------------------------------------------------------===//
///
/// \file
/// This file contains the declaration of the LLVM code pattern finder and
/// comparison manager. The comparator supports both instruction-based and
/// value-based difference patterns.
///
//===----------------------------------------------------------------------===//

#ifndef DIFFKEMP_SIMPLL_CUSTOMPATTERNCOMPARATOR_H
#define DIFFKEMP_SIMPLL_CUSTOMPATTERNCOMPARATOR_H

#include "CustomPatternSet.h"
#include "InstPatternComparator.h"
#include "ValuePatternComparator.h"
#include <llvm/IR/Instructions.h>
#include <unordered_map>
#include <vector>

using namespace llvm;

/// Compares difference patterns against functions, possibly eliminating reports
/// of prior semantic differences.
/// Supported difference patterns:
///   - Instruction patterns - Patterns consisting of standard LLVM IR code,
///     i.e., typically containing multiple instructions and basic blocks.
///     Instructions may be annotated using diffkemp.pattern metadata.
///   - Value patterns - Patterns describing a change of a single value. The
///     original and the modified value are described strictly through return
///     instructions.
/// Concrete examples of difference patterns can be seen in regression tests.
class CustomPatternComparator {
  public:
    CustomPatternComparator(
            const CustomPatternSet *Patterns,
            const DifferentialFunctionComparator *DiffFunctionComp,
            const Function *FnL,
            const Function *FnR)
            : DiffFunctionComp(DiffFunctionComp) {
        addPatternSet(Patterns, FnL, FnR);
    };

    /// Individual matched instructions, combined for all matched patterns.
    /// Combines results of all active pattern matches.
    InstructionSet AllInstMatches;
    /// Mapping for matched module instructions. Instructions from the left
    /// module are used as keys, while instructions from the right module
    /// are used as values. Contains the result of a single pattern match.
    mutable InstructionMap InstMappings;

    /// Tries to match a difference pattern starting with the given instruction
    /// pair. Returns true if a valid match is found.
    bool matchPattern(const Instruction *InstL, const Instruction *InstR);

    /// Tries to match a pair of values to a value pattern. Returns true if a
    /// valid match is found.
    bool matchValues(const Value *L, const Value *R);

    /// Populate the comparator maps with a given set of patterns.
    void addPatternSet(const CustomPatternSet *PatternSet,
                       const Function *FnL,
                       const Function *FnR);

  private:
    /// Pair of corresponding instruction pattern function comparators.
    using InstPatternComparatorPair =
            std::pair<std::unique_ptr<InstPatternComparator>,
                      std::unique_ptr<InstPatternComparator>>;
    /// Pair of corresponding value pattern function comparators.
    using ValuePatternComparatorPair =
            std::pair<std::unique_ptr<ValuePatternComparator>,
                      std::unique_ptr<ValuePatternComparator>>;

    /// Parent DifferentialFunctionComparator the compares the module functions.
    const DifferentialFunctionComparator *DiffFunctionComp;
    /// Instruction pattern function comparators associated with the current set
    /// of patterns and the currently compared functions.
    DenseMap<const InstPattern *, InstPatternComparatorPair> InstPatternComps;
    /// Value pattern function comparators associated with the current set of
    /// patterns and the currently compared functions.
    DenseMap<const ValuePattern *, ValuePatternComparatorPair>
            ValuePatternComps;

    /// Add an instruction pattern to the instruction pattern comparator map.
    void addInstPattern(const InstPattern &Pattern,
                        const Function *FnL,
                        const Function *FnR);

    /// Add a value pattern to the value pattern comparator map.
    void addValuePattern(const ValuePattern &Pattern,
                         const Function *FnL,
                         const Function *FnR);

    /// Tries to match one of the loaded instruction patterns. Returns true
    /// if a valid match is found.
    bool matchInstPattern(const Instruction *InstL, const Instruction *InstR);

    /// Tries to match one of the loaded value patterns. Returns true if
    /// a valid match is found.
    bool matchValuePattern(const Instruction *InstL, const Instruction *InstR);

    /// Tries to match a load instruction to the start of the given value
    /// pattern.
    bool matchLoadInst(const LoadInst *Load,
                       const ValuePattern *Pat,
                       bool IsLeftSide);

    /// Check whether the input mapping generated by the given pattern function
    /// comparator pair is valid even when both compared modules are analysed at
    /// once.
    bool inputMappingValid(const InstPattern *Pat,
                           const InstPatternComparatorPair *PatternComps);

    /// Create the resulting instruction mapping and add all matched
    /// instructions into the combined instruction set.
    void processPatternMatch(const InstPattern *Pat,
                             const InstPatternComparatorPair *PatternComps);
};

#endif // DIFFKEMP_SIMPLL_CUSTOMPATTERNCOMPARATOR_H
